/*
 * Kuklomenos
 * Copyright (C) 2008-2009 Martin Bays <mbays@sdf.lonestar.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 */

#include <config.h>
#ifdef HAVE_LIBCURL
#define HIGH_SCORE_REPORTING 1
#endif

#include <string>
#include <map>
using namespace std;

#include "keybindings.h"

const command C_FIRST = C_LEFT;
const command C_LAST = C_SCREENSHOT;
const command C_LASTDEBUG = C_WIN;

string longCommandNames[] = {
    "BUG",
    "Turn left",
    "Turn right",
    "Dezoom",
    "De-aim",
    "Shoot green",
    "Shoot yellow",
    "Shoot red",
    "Shoot pod",
    "Pause",
    "Quit",
    "Menu",
    "Menu left",
    "Menu up",
    "Menu right",
    "Menu down",
    "Start game",
#ifdef SOUND
    "Toggle sound",
#endif
    "Toggle zoom",
    "Toggle rotate",
    "Toggle grid",
    "Decrease anti-aliasing",
    "Increase anti-aliasing",
    "Decrease fps",
    "Increase fps",
    "Decrease game rate",
    "Increase game rate",
    "Reset game rate",
#ifdef HIGH_SCORE_REPORTING
    "Report high score",
#endif
    "Take screenshot",
    "[DEBUG] Invulnerability",
    "[DEBUG] Win",
};
string shortCommandNames[] = {
    "BUG",
    "left",
    "right",
    "dezoom",
    "deaim",
    "shoot1",
    "shoot2",
    "shoot3",
    "shoot4",
    "pause",
    "quit",
    "menu",
    "menu_left",
    "menu_up",
    "menu_right",
    "menu_down",
    "start",
#ifdef SOUND
    "sound",
#endif
    "zoom",
    "rotate",
    "grid",
    "decaa",
    "incaa",
    "decfps",
    "incfps",
    "decrate",
    "incrate",
    "resetrate",
#ifdef HIGH_SCORE_REPORTING
    "reporths",
#endif
    "screenshot",
    "invuln",
    "win"
};

command commandOfString(const string& str)
{
    for (command c = C_FIRST; c <= C_LASTDEBUG; c = command(c+1))
	if (str == shortCommandNames[c])
	    return c;
    return C_NONE;
}

bool isMenuCommand(const command& c)
{
    return (c >= C_M_LEFT && c <= C_M_DOWN);
}

SDLMod Key::simpleMod(const SDLMod& in) const
{
    return SDLMod(((in & KMOD_SHIFT) ? KMOD_SHIFT : KMOD_NONE) |
	    ((in & KMOD_CTRL) ? KMOD_CTRL : KMOD_NONE) |
	    (((in & KMOD_ALT) | (in & KMOD_META)) ?
	     KMOD_ALT : KMOD_NONE));
}

Key::Key(SDL_keysym keysym)
{
    // care only about shift, ctrl, and alt/meta.
    sym = keysym.sym;
    mod = simpleMod(keysym.mod);
}

struct strcomp {
    bool operator() (const string& str1, const string& str2) const
    { return (str1.compare(str2) < 0); }
};

Key::Key(string keyString)
{
    static map<string,SDLKey,strcomp> keynameMap;
    if (keynameMap.size() == 0)
    {
	// initialise keynameMap
	for (SDLKey k = SDLK_FIRST; k != SDLK_LAST; k = SDLKey(k+1))
	{
	    string s = SDL_GetKeyName(k);
	    keynameMap.insert( pair<string,SDLKey>(s,k) );
	}
    }

    mod = KMOD_NONE;
    size_t oldlen;
    do {
	oldlen = keyString.length();
	if (keyString.compare(0, 6, "shift+") == 0)
	{ keyString.erase(0, 6); mod = SDLMod(mod | KMOD_SHIFT); }
	if (keyString.compare(0, 5, "ctrl+") == 0)
	{ keyString.erase(0, 5); mod = SDLMod(mod | KMOD_CTRL); }
	if (keyString.compare(0, 4, "alt+") == 0)
	{ keyString.erase(0, 4); mod = SDLMod(mod | KMOD_ALT); }
    } while (keyString.length() < oldlen);

    map<string,SDLKey,strcomp>::iterator it = keynameMap.find(keyString);
    if ( it != keynameMap.end() )
	sym = it->second;
    else
	sym = SDLK_UNKNOWN;
}

string Key::getString() const
{
    if (sym == SDLK_UNKNOWN)
	return string("none");

    string s("");

    if (mod & KMOD_SHIFT)
	s += string("shift+");
    if (mod & KMOD_CTRL)
	s += string("ctrl+");
    if (mod & KMOD_ALT)
	s += string("alt+");
    s += string(SDL_GetKeyName(sym));

    return s;
}

bool Key::operator!=(Key const& other) const
{
    return (!((*this) == other));
}
bool Key::operator==(Key const& other) const
{
    return (sym == other.sym && mod == other.mod);
}

bool Key::isPressed()
{
    return (SDL_GetKeyState(NULL)[sym] && simpleMod(SDL_GetModState()) == mod);
}

Key KEY_NONE;

// get(command): returns Key bound to command if in the map, else the null Key
Key Keybindings::get(command com) const
{
    map<command,Key>::const_iterator it = find(com);
    if (it == end())
	return Key();
    else
	return it->second;
}

// XXX: see http://www.parashift.com/c++-faq-lite/ctors.html#faq-10.12 if
// you're wondering why we don't just make defaultKeybindings an extern
Keybindings& defaultKeybindings()
{
    static Keybindings defaultKeybindings;
    if (defaultKeybindings.empty())
    {
	//initialise
	defaultKeybindings[C_LEFT] = Key(SDLK_LEFT);
	defaultKeybindings[C_RIGHT] = Key(SDLK_RIGHT);
	defaultKeybindings[C_DEZOOM] = Key(SDLK_UP);
	defaultKeybindings[C_DEAIM] = Key(SDLK_DOWN);
	defaultKeybindings[C_SHOOT_GREEN] = Key(SDLK_1);
	defaultKeybindings[C_SHOOT_YELLOW] = Key(SDLK_2);
	defaultKeybindings[C_SHOOT_RED] = Key(SDLK_3);
	defaultKeybindings[C_SHOOT_POD] = Key(SDLK_4);

	defaultKeybindings[C_PAUSE] = Key(SDLK_p);
	defaultKeybindings[C_QUIT] = Key(SDLK_q);

	defaultKeybindings[C_MENU] = Key(SDLK_ESCAPE);
	defaultKeybindings[C_M_LEFT] = Key(SDLK_LEFT);
	defaultKeybindings[C_M_UP] = Key(SDLK_UP);
	defaultKeybindings[C_M_RIGHT] = Key(SDLK_RIGHT);
	defaultKeybindings[C_M_DOWN] = Key(SDLK_DOWN);

	defaultKeybindings[C_STARTGAME] = Key(SDLK_SPACE);

#ifdef SOUND
	defaultKeybindings[C_SOUND] = Key(SDLK_s);
#endif

	defaultKeybindings[C_ZOOM] = Key(SDLK_z);
	defaultKeybindings[C_ROTATE] = Key(SDLK_r);
	defaultKeybindings[C_GRID] = Key(SDLK_g);
	defaultKeybindings[C_DECAA] = Key(SDLK_a, SDLMod(KMOD_SHIFT));
	defaultKeybindings[C_INCAA] = Key(SDLK_a);

	defaultKeybindings[C_DECFPS] = Key(SDLK_LEFTBRACKET);
	defaultKeybindings[C_INCFPS] = Key(SDLK_RIGHTBRACKET);

	defaultKeybindings[C_DECRATE] = Key(SDLK_MINUS);
	defaultKeybindings[C_INCRATE] = Key(SDLK_EQUALS, SDLMod(KMOD_SHIFT));
	defaultKeybindings[C_RESETRATE] = Key(SDLK_EQUALS);

#ifdef HIGH_SCORE_REPORTING
	defaultKeybindings[C_REPORTHS] = Key(SDLK_F10);
#endif

	defaultKeybindings[C_SCREENSHOT] = Key(SDLK_F12);

	defaultKeybindings[C_INVULN] = Key(SDLK_F4);
	defaultKeybindings[C_WIN] = Key(SDLK_F5);
    }

    return defaultKeybindings;
}
